#!/usr/bin/python
'''
Created on 23 Sep 2017

@author: kamichal
'''

from contextlib import contextmanager
import os
import yattag

from .six import str_types
from .utils import fix_yattag, dictionize_css, form_css, assert_keys_not_in, make_place
from yawrap._formatter import HtmlFormatter


DEFAULT_SVG_TAG_ATTRIBUTES = dict(xmlns="http://www.w3.org/2000/svg", version="1.1")

fix_yattag(yattag)


@contextmanager
def svg_structure(painter,
                  svg_tag_attributes=DEFAULT_SVG_TAG_ATTRIBUTES,
                  svg_styles_as_str="",
                  *args,
                  **kwargs):

    kwargs.update(svg_tag_attributes)
    svg_defs = None
    with painter.tag('svg', *args, **kwargs):
        if svg_styles_as_str:
            with painter.tag('style', type="text/css"):
                painter.cdata(svg_styles_as_str)
        if svg_defs:
            with painter.tag('defs', type="text/css"):
                for id_, svg_def in svg_defs.items():
                    with painter.tag('g', id=id_):
                        painter.asis(svg_def)
        yield


class Yawrap(yattag.Doc):
    css = ''
    js = []
    linked_js = []
    linked_css = []
    html_d = dict(lang="en-US")
    meta_d = [dict(charset="UTF-8")]
    svg_d = DEFAULT_SVG_TAG_ATTRIBUTES
    html_formatter = HtmlFormatter.yattag_indent

    def __init__(self, target_file, title='', parent=None, defaults=None, errors=None,
                 error_wrapper=('<span class="error">', '</span>'), stag_end='/>'):

        super(Yawrap, self).__init__(defaults, errors, error_wrapper, stag_end)
        self._target_file = target_file
        self.title = title
        self._parent = parent
        self._target_dir = os.path.dirname(target_file)
        self._additional_js = []
        self._additional_linked_js = []
        self._additional_css = {}
        self.external_csss = set()

    @contextmanager
    def local_link(self, target, *args, **kwargs):
        rel_location = os.path.relpath(target, self._target_dir)
        assert_keys_not_in('href', args, kwargs)
        with self.tag('a', href=rel_location, *args, **kwargs):
            yield rel_location

    def render(self):
        """ Saves page in current state to target file. """
        raw_html = self._render_page()
        with open(make_place(self._target_file), 'wt') as ff:
            ff.write(raw_html)

    @contextmanager
    def svg(self, svg_styles_as_str='', *args, **svg_tag_attributes):
        svg_tag_attributes.update(self.svg_d)
        with svg_structure(self, svg_tag_attributes, svg_styles_as_str, *args, **svg_tag_attributes):
            yield

    def add_css(self, css_rules):
        if not isinstance(css_rules, dict):
            assert isinstance(css_rules, str_types)
            css_rules = dictionize_css(css_rules)
        self._additional_css.update(css_rules)

    def add_js(self, js_script):
        assert isinstance(js_script, str_types)
        self._additional_js.append(js_script)

    def _get_rel_path(self, target_local_file):
        return os.path.relpath(os.path.abspath(target_local_file), self._target_dir)

    def link_local_css_file(self, target_css_file_path):
        self.external_csss.add(self._get_rel_path(target_css_file_path))

    def link_local_js_file(self, target_js_file_path):
        self._additional_linked_js.add(self._get_rel_path(target_js_file_path))

    def link_external_js_file(self, target_js_url):
        self._additional_linked_js.append(target_js_url)

    def link_external_css_file(self, target_css_url):
        self.external_csss.add(target_css_url)

    @contextmanager
    def _html_page_structure(self, page_doc):
        page_css = dictionize_css(self.css).copy()
        page_css.update(self._additional_css)
        page_js = self._additional_js + self.js
        linked_js = self._additional_linked_js + self.linked_js

        page_doc.asis('<!doctype html>')
        with page_doc.tag('html', **self.html_d):
            with page_doc.tag('head'):
                for meta in self.meta_d:
                    page_doc.stag('meta', **meta)
                if self.title:
                    with page_doc.tag('title'):
                        page_doc.text(self.title)
                for js_link in linked_js:
                    with page_doc.tag("script", src=js_link):
                        pass
                for js in page_js:
                    with page_doc.tag('script'):
                        page_doc.asis(js)
                if page_css:
                    with page_doc.tag('style'):
                        page_doc.asis(form_css(page_css, indent_level=3))
                for ext_css in self.external_csss:
                    page_doc.stag("link", rel="stylesheet", type="text/css", href=ext_css)
            with page_doc.tag('body'):
                yield

    def _get_body_render(self):
        raw_body_html = self.getvalue()
        return self.html_formatter(raw_body_html)

    def _render_page(self):
        page_doc = yattag.SimpleDoc()
        with self._html_page_structure(page_doc):
            page_doc.asis(self._get_body_render())

        raw_text = page_doc.getvalue()
        return self.html_formatter(raw_text)




import yattag

from .utils import fix_yattag
from .elements_def import H4, H5, CONTEXT_ELEMENTS, EMPTY_ELEMENTS

fix_yattag(yattag)


class Ya5(yattag.SimpleDoc):
    version = H5

    def __init__(self, stag_end=">"):
        super(Ya5, self).__init__(stag_end)

        def create_contextual(elm):
            def _tag(*args, **kwargs):
                return self.tag(elm.tag, *args, **kwargs)
            _tag.__name__ = elm.tag
            _tag.__doc__ = elm.info

            return _tag

        def create_empty(elm):
            def _stag(*args, **kwargs):
                return self.stag(elm.tag, *args, **kwargs)
            _stag.__name__ = elm.tag
            _stag.__doc__ = elm.info

            return _stag

        for elem in CONTEXT_ELEMENTS[self.version]:
            setattr(self, elem.tag, create_contextual(elem))

        for elem in EMPTY_ELEMENTS[self.version]:
            setattr(self, elem.tag, create_empty(elem))


class Ya4(Ya5):
    version = H4



#!/usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 30 wrz 2017

@author: kamichal
'''
import os
import re
from yattag.simpledoc import ATTR_NO_VALUE

from .six import str_types

KNOWN_SUBSTITUTES = {
    'klass': 'class',
    'Class': 'class',
    'class_': 'class',
    'fill_opacity': 'fill-opacity',
    'stroke_width': 'stroke-width',
    'stroke_dasharray': ' stroke-dasharray',
    "stroke_opacity": "stroke-opacity",
    "stroke_dashoffset": "stroke-dashoffset",
    "stroke_linejoin": "stroke-linejoin",
    "stroke_miterlimit": "stroke-miterlimit",
}


def _attributes2(args, kwargs):

    def tr(arg):
        if isinstance(arg, tuple):
            return arg
        elif isinstance(arg, str_types):
            return (arg, ATTR_NO_VALUE)
        else:
            raise ValueError("Couldn't make a XML or HTML attribute/value pair out of %s." % repr(arg))

    result = dict(map(tr, args))
    result.update({KNOWN_SUBSTITUTES.get(k, k): v for k, v in kwargs.items()})
    return result


def fix_yattag(yattag_module):
    yattag_module.simpledoc._attributes = _attributes2


def make_place(target_file):
    dir_ = os.path.dirname(target_file)
    if not os.path.isdir(dir_):
        assert not os.path.isfile(dir_)
        os.makedirs(dir_)
    return target_file


def assert_keys_not_in(keys, args, kwargs):
    keys = keys if isinstance(keys, (list, tuple)) else [keys]
    for key in keys:
        defined_keys = list(kwargs.keys()) + [x[0] for x in args if isinstance(x, tuple)]
        if key in defined_keys:
            raise ValueError("Duplicated '{}' attribute.".format(key))


def dictionize_css(input_css):
    if isinstance(input_css, dict):
        return input_css
    assert isinstance(input_css, str_types)
    input_css = input_css.strip()
    if input_css and any(bracket not in input_css for bracket in ("{", "}")):
        raise ValueError("That's most probably not a CSS code: {}".format(input_css))

    comment_prog = re.compile(r"(\/\*.*\*\/)")
    input_css = comment_prog.sub("", input_css)

    def iterate_rules(in_css):
        while in_css:
            decl_start = in_css.find('{')
            decl_end = in_css.find('}')
            if any(o == -1 for o in (decl_start, decl_end)):
                break
            selectors = in_css[0:decl_start].strip()
            declarations = in_css[decl_start + 1: decl_end].strip()
            yield selectors, declarations
            in_css = in_css[decl_end + 1:]

    def iterate_declarations(declarations):
        for decl in declarations.split(';'):
            if decl:
                first_colon = decl.find(':')
                if first_colon:
                    property_ = decl[:first_colon].strip()
                    value = decl[first_colon + 1:].strip()
                    yield property_, value
                else:
                    yield decl.strip(), None

    return {selector: {prop: val for prop, val in iterate_declarations(declarations)}
            for selector, declarations in iterate_rules(input_css)}


def form_css(structured_css, indent_level=1):
    assert isinstance(structured_css, dict)
    base_indent = ' ' * 2
    indent = base_indent * indent_level

    template = "\n{ind}{selector} {{{definitions}}}"
    def_tpl = "{ind}{bind}{property}: {value};"

    def rules():
        for selector, definitions in sorted(structured_css.items()):
            defs = '\n'.join(def_tpl.format(ind=indent, bind=base_indent, property=prop, value=val)
                             for prop, val in sorted(definitions.items()))
            if defs:
                defs = "\n{}\n{ind}".format(defs, ind=indent)
            yield template.format(ind=indent, selector=selector, definitions=defs)

    return ''.join(rules())



import re
from collections import namedtuple

_INDENT_PROG = re.compile("\s*\>\n\s*\<")
_END_TAG_PROG = re.compile("(\</\s*\w+\s*\>)")


@staticmethod
def _no_indent(raw_html_string):
    """ no indentation,
        the fastest store (no html postprocessing),
        slow read and JS DOM traverse,
        smallest file but unreadable for a human """
    if "\n" in raw_html_string:
        return _INDENT_PROG.sub("><", raw_html_string)
    return raw_html_string


@staticmethod
def _limited_line_width(raw_html_string):
    """ no indentation,
        ugly but fast store (html render),
        fast read and JS DOM traverse
        acceptable file size
    """
    max_line_width = 240

    def chop():
        cache, len_ = [], 0
        for part in raw_html_string.split(">"):
            cache.append(part)
            len_ += 1 + len(part)
            if len_ > max_line_width:
                yield ">".join(cache)
                cache, len_ = [], 0
        if cache:
            yield ">".join(cache)

    return ">\n".join(chop())


@staticmethod
def _new_line_each_end(raw_html_string):
    """ no indentation, new line at each ending tag
        fast store (html render) fast read and JS DOM traverse
        larger file size
    """
    return _END_TAG_PROG.sub(r"\1\n", raw_html_string)


@staticmethod
def _yattag_indent(raw_html_string):
    """ the prettiest html, but slow store (because of html tokenization),
        largest file size, fast read and JS DOM traverse
    """
    from yattag import indent
    return indent(raw_html_string)


# It's assumed they are sorted in order of performance
_HtmlFormatterOption = namedtuple("_HtmlFormatterOption", [
    "no_indent",
    "limited_line_width",
    "new_line_each_end",
    "yattag_indent",
])

HtmlFormatter = _HtmlFormatterOption(
    no_indent=_no_indent,
    limited_line_width=_limited_line_width,
    new_line_each_end=_new_line_each_end,
    yattag_indent=_yattag_indent,
)



from collections import namedtuple

H4 = 'html4'  # only in html4
H5 = 'html5'  # only in html5
HB = 'both'  # allowed in both
N = 'normal'  # has a closing tag
E = 'empty'  # doesn't have a closing tag

ES = namedtuple('ES', 'tag, standard, type_, info')
ER = namedtuple('ES', 'tag, info')

HTML_ELEMENTS = [
    ES("html", HB, N, "Defines an HTML document"),
    ES("head", HB, N, "Defines information about the document"),
    ES("title", HB, N, "Defines a title for the document"),
    ES("body", HB, N, "Defines the document's body"),
    ES("h1", HB, N, "Defines HTML heading, rank 1."),
    ES("h2", HB, N, "Defines HTML heading, rank 2."),
    ES("h3", HB, N, "Defines HTML heading, rank 3."),
    ES("h4", HB, N, "Defines HTML heading, rank 4."),
    ES("h5", HB, N, "Defines HTML heading, rank 5."),
    ES("h6", HB, N, "Defines HTML heading, rank 6."),
    ES("p", HB, N, "Defines a paragraph"),
    ES("br", HB, E, "Inserts a single line break"),
    ES("hr", HB, E, "Defines a thematic change in the content"),
    # Formatting
    ES("acronym", H4, N, "Not supported in HTML5. Use <abbr> instead. Defines an acronym."),
    ES("abbr", HB, N, "Defines an abbreviation or an acronym"),
    ES("address", HB, N, "Defines contact information for the author/owner of a document/article"),
    ES("b", HB, N, "Defines bold text"),
    ES("bdi", H5, N, "Isolates a part of text that might be formatted in a different direction from other text outside it"),
    ES("bdo", HB, N, "Overrides the current text direction"),
    ES("big", H4, N, "Not supported in HTML5. Use CSS instead. Defines big text"),
    ES("blockquote", HB, N, "Defines a section that is quoted from another source"),
    ES("center", H4, N, "Not supported in HTML5. Use CSS instead. Defines centered text."),
    ES("cite", HB, N, "Defines the title of a work"),
    ES("code", HB, N, "Defines a piece of computer code"),
    ES("del", HB, N, "Defines text that has been deleted from a document"),
    ES("dfn", HB, N, "Represents the defining instance of a term"),
    ES("em", HB, N, "Defines emphasized text "),
    ES("font", H4, N, "Not supported in HTML5. Use CSS instead. Defines font, color, and size for text."),
    ES("i", HB, N, "Defines a part of text in an alternate voice or mood"),
    ES("ins", HB, N, "Defines a text that has been inserted into a document"),
    ES("kbd", HB, N, "Defines keyboard input"),
    ES("mark", H5, N, "Defines marked/highlighted text"),
    ES("meter", HB, N, "Defines a scalar measurement within a known range (a gauge)"),
    ES("pre", HB, N, "Defines preformatted text"),
    ES("progress", H5, N, "Represents the progress of a task"),
    ES("q", HB, N, "Defines a short quotation"),
    ES("rp", H5, N, "Defines what to show in browsers that do not support ruby annotations"),
    ES("rt", H5, N, "Defines an explanation/pronunciation of characters (for East Asian typography)"),
    ES("ruby", H5, N, "Defines a ruby annotation (for East Asian typography)"),
    ES("s", HB, N, "Defines text that is no longer correct"),
    ES("samp", HB, N, "Defines sample output from a computer program"),
    ES("small", HB, N, "Defines smaller text"),
    ES("strike", H4, N, "Not supported in HTML5. Use <del> or <s> instead. Defines strikethrough text"),
    ES("strong", HB, N, "Defines important text"),
    ES("sub", HB, N, "Defines subscripted text"),
    ES("sup", HB, N, "Defines superscripted text"),
    ES("time", H5, N, "Defines a date/time"),
    ES("tt", H4, N, "Not supported in HTML5. Use CSS instead. Defines teletype text."),
    ES("u", HB, N, "Defines text that should be stylistically different from normal text"),
    ES("var", HB, N, "Defines a variable"),
    ES("wbr", H5, E, "Defines a possible line-break"),
    # Forms and Input
    ES("form", HB, N, "Defines an HTML form for user input"),
    ES("input", HB, E, "Defines an input control"),
    ES("textarea", HB, N, "Defines a multiline input control (text area)"),
    ES("button", HB, N, "Defines a clickable button"),
    ES("select", HB, N, "Defines a drop-down list"),
    ES("optgroup", HB, N, "Defines a group of related options in a drop-down list"),
    ES("option", HB, N, "Defines an option in a drop-down list"),
    ES("label", HB, N, "Defines a label for an <input> element"),
    ES("fieldset", HB, N, "Groups related elements in a form"),
    ES("legend", HB, N, "Defines a caption for a <fieldset> element"),
    ES("datalist", H5, N, "Specifies a list of pre-defined options for input controls"),
    ES("keygen", H5, E, "Defines a key-pair generator field (for forms)"),
    ES("output", H5, N, "Defines the result of a calculation"),
    # Frames
    ES("frame", H4, N, "Not supported in HTML5. Defines a window (a frame) in a frameset"),
    ES("frameset", H4, N, "Not supported in HTML5. Defines a set of frames"),
    ES("noframes", H4, N, "Not supported in HTML5. Defines an alternate content for users that do not support frames"),
    ES("iframe", HB, N, "Defines an inline frame"),
    # Images
    ES("img", HB, E, "Defines an image"),
    ES("map", HB, N, "Defines a client-side image-map"),
    ES("area", HB, E, "Defines an area inside an image-map"),
    ES("canvas", H5, N, "Used to draw graphics, on the fly, via scripting (usually JavaScript)"),
    ES("figcaption", H5, N, "Defines a caption for a <figure> element"),
    ES("figure", H5, N, "Specifies self-contained content"),
    ES("picture", H5, N, "Defines a container for multiple image resources"),
    # Audio / Video
    ES("audio", H5, N, "Defines sound content"),
    ES("source", H5, E, "Defines multiple media resources for media elements (<video>, <audio> and <picture>"),
    ES("track", H5, E, "Defines text tracks for media elements (<video> and <audio>"),
    ES("video", H5, N, "Defines a video or movie"),
    # Links
    ES("a", HB, N, "Defines a hyperlink"),
    ES("link", HB, E, "Defines the relationship between a document and an external resource (most used to link to style sheets)"),
    ES("nav", H5, N, "Defines navigation links"),
    # Lists
    ES("ul", HB, N, "Defines an unordered list"),
    ES("ol", HB, N, "Defines an ordered list"),
    ES("li", HB, N, "Defines a list item"),
    ES("dir", H4, N, "Not supported in HTML5. Use <ul> instead. Defines a directory list"),
    ES("dl", HB, N, "Defines a description list"),
    ES("dt", HB, N, "Defines a term/name in a description list"),
    ES("dd", HB, N, "Defines a description of a term/name in a description list"),
    ES("menu", HB, N, "Defines a list/menu of commands"),
    ES("menuitem", H5, N, "Defines a command/menu item that the user can invoke from a popup menu"),
    # Tables
    ES("table", HB, N, "Defines a table"),
    ES("caption", HB, N, "Defines a table caption"),
    ES("th", HB, N, "Defines a header cell in a table"),
    ES("tr", HB, N, "Defines a row in a table"),
    ES("td", HB, N, "Defines a cell in a table"),
    ES("thead", HB, N, "Groups the header content in a table"),
    ES("tbody", HB, N, "Groups the body content in a table"),
    ES("tfoot", HB, N, "Groups the footer content in a table"),
    ES("col", HB, E, "Specifies column properties for each column within a <colgroup> element"),
    ES("colgroup", HB, N, "Specifies a group of one or more columns in a table for formatting"),
    # Styles and Semantics
    ES("style", HB, N, "Defines style information for a document"),
    ES("div", HB, N, "Defines a section in a document"),
    ES("span", HB, N, "Defines a section in a document"),
    ES("header", H5, N, "Defines a header for a document or section"),
    ES("footer", H5, N, "Defines a footer for a document or section"),
    ES("main", H5, N, "Specifies the main content of a document"),
    ES("section", H5, N, "Defines a section in a document"),
    ES("article", H5, N, "Defines an article"),
    ES("aside", H5, N, "Defines content aside from the page content"),
    ES("details", H5, N, "Defines additional details that the user can view or hide"),
    ES("dialog", H5, N, "Defines a dialog box or window"),
    ES("summary", H5, N, "Defines a visible heading for a <details> element"),
    ES("data", H5, N, "Links the given content with a machine-readable translation"),
    # Meta Info
    ES("head", HB, N, "Defines information about the document"),
    ES("meta", HB, E, "Defines metadata about an HTML document"),
    ES("base", HB, E, "Specifies the base URL/target for all relative URLs in a document"),
    ES("basefont", H4, N, "Not supported in HTML5. Use CSS instead. Specifies a default color, size, and font for all text in a document"),
    # Programming
    ES("script", HB, N, "Defines a client-side script"),
    ES("noscript", HB, N, "Defines an alternate content for users that do not support client-side scripts"),
    ES("applet", H4, N, "Not supported in HTML5. Use <embed> or <object> instead. Defines an embedded applet"),
    ES("embed", H5, E, "Defines a container for an external (non-HTML) application"),
    ES("object", HB, N, "Defines an embedded object"),
    ES("param", HB, E, "Defines a parameter for an object")]

HTML4_ELEMENTS = [e for e in HTML_ELEMENTS if e.standard == H4 or e.standard == HB]
HTML5_ELEMENTS = [e for e in HTML_ELEMENTS if e.standard == H5 or e.standard == HB]

HTML_TAGS = {
    H4: [e.tag for e in HTML4_ELEMENTS],
    H5: [e.tag for e in HTML5_ELEMENTS],
}
CONTEXT_ELEMENTS = {
    H4: [ER(e.tag, e.info) for e in HTML4_ELEMENTS if e.type_ == N],
    H5: [ER(e.tag, e.info) for e in HTML5_ELEMENTS if e.type_ == N]
}

EMPTY_ELEMENTS = {
    H4: [ER(e.tag, e.info) for e in HTML4_ELEMENTS if e.type_ == E],
    H5: [ER(e.tag, e.info) for e in HTML5_ELEMENTS if e.type_ == E]
}
